async function visitObject(objectIn, handleNode) {
  const objectOut = {}
  const objectKeys = Object.keys(objectIn)
  for (let i = 0; i < objectKeys.length; i++) {
    const key = objectKeys[i]
    objectOut[key] = await visitProperty(key, objectIn[key], handleNode)
  }
  return objectOut
}

async function visitProperty(key, value, handleNode) {
  const handledValue = await handleNode(key, value)
  if (isObject(value)) {
    return await visitObject(handledValue, handleNode)
  } else if (isArray(handledValue)) {
    return await Promise.all(
      handledValue.map((v) => visitProperty(key, v, handleNode))
    )
  } else {
    return handledValue
  }
}

// Returns true if a value is an array, false otherwise
export function isArray(value: unknown): value is any[] {
  return Array.isArray(value)
}

// Returns true if a value is an object, false otherwise
export function isObject(value: unknown): value is Record<string, unknown> {
  return value && typeof value === 'object' && value.constructor === Object
}

export async function traverse(valueIn, handleNode = (k, v) => v) {
  return await visitProperty(false, valueIn, handleNode)
}
